@@ -9,18 +9,22 @@ from click import UsageError
 import mlflow.db
 import mlflow.experiments
 import mlflow.deployments.cli
+import mlflow.pipelines.cli
 import mlflow.projects as projects
 import mlflow.runs
 import mlflow.store.artifact.cli
+from mlflow import version
 from mlflow import tracking
-from mlflow.store.tracking import DEFAULT_LOCAL_FILE_AND_ARTIFACT_PATH
+from mlflow.store.tracking import DEFAULT_LOCAL_FILE_AND_ARTIFACT_PATH, DEFAULT_ARTIFACTS_URI
 from mlflow.store.artifact.artifact_repository_registry import get_artifact_repository
 from mlflow.tracking import _get_store
 from mlflow.utils import cli_args
-from mlflow.utils.annotations import experimental
 from mlflow.utils.logging_utils import eprint
 from mlflow.utils.process import ShellCommandException
-from mlflow.utils.uri import is_local_uri
+from mlflow.utils.server_cli_utils import (
+    resolve_default_artifact_root,
+    artifacts_only_config_validation,
+)
 from mlflow.entities.lifecycle_stage import LifecycleStage
 from mlflow.exceptions import MlflowException
 
@@ -28,7 +32,7 @@ _logger = logging.getLogger(__name__)
 
 
 @click.group()
-@click.version_option()
+@click.version_option(version=version.VERSION)
 def cli():
     pass
 
@@ -107,6 +111,7 @@ def cli():
     "at https://www.mlflow.org/docs/latest/projects.html.",
 )
 @cli_args.NO_CONDA
+@cli_args.ENV_MANAGER
 @click.option(
     "--storage-dir",
     envvar="MLFLOW_TMP_DIR",
@@ -121,6 +126,12 @@ def cli():
     "Note: this argument is used internally by the MLflow project APIs "
     "and should not be specified.",
 )
+@click.option(
+    "--run-name",
+    metavar="RUN_NAME",
+    help="The name to give the MLflow Run associated with the project execution. If not specified, "
+    "the MLflow Run name is left unset.",
+)
 def run(
     uri,
     entry_point,
@@ -131,9 +142,11 @@ def run(
     experiment_id,
     backend,
     backend_config,
-    no_conda,
+    no_conda,  # pylint: disable=unused-argument
+    env_manager,
     storage_dir,
     run_id,
+    run_name,
 ):
     """
     Run an MLflow project from the given URI.
@@ -175,10 +188,11 @@ def run(
             docker_args=args_dict,
             backend=backend,
             backend_config=backend_config,
-            use_conda=(not no_conda),
+            env_manager=env_manager,
             storage_dir=storage_dir,
             synchronous=backend in ("local", "kubernetes") or backend is None,
             run_id=run_id,
+            run_name=run_name,
         )
     except projects.ExecutionException as e:
         _logger.error("=== %s ===", e)
@@ -233,25 +247,32 @@ def _validate_server_args(gunicorn_opts=None, workers=None, waitress_opts=None):
     "SQLAlchemy-compatible database connection strings "
     "(e.g. 'sqlite:///path/to/file.db') or local filesystem URIs "
     "(e.g. 'file:///absolute/path/to/directory'). By default, data will be logged "
-    "to the ./mlruns directory.",
+    f"to {DEFAULT_LOCAL_FILE_AND_ARTIFACT_PATH}",
 )
 @click.option(
     "--default-artifact-root",
     metavar="URI",
     default=None,
-    help="Path to local directory to store artifacts, for new experiments. "
-    "Note that this flag does not impact already-created experiments. "
-    "Default: " + DEFAULT_LOCAL_FILE_AND_ARTIFACT_PATH,
+    help="Directory in which to store artifacts for any new experiments created. For tracking "
+    "server backends that rely on SQL, this option is required in order to store artifacts. "
+    "Note that this flag does not impact already-created experiments with any previous "
+    "configuration of an MLflow server instance. "
+    "If the --serve-artifacts option is specified, the default artifact root is "
+    f"{DEFAULT_ARTIFACTS_URI}. Otherwise, the default artifact root is "
+    f"{DEFAULT_LOCAL_FILE_AND_ARTIFACT_PATH}.",
 )
+@cli_args.SERVE_ARTIFACTS
 @cli_args.ARTIFACTS_DESTINATION
 @cli_args.PORT
 @cli_args.HOST
-def ui(backend_store_uri, default_artifact_root, artifacts_destination, port, host):
+def ui(
+    backend_store_uri, default_artifact_root, serve_artifacts, artifacts_destination, port, host
+):
     """
     Launch the MLflow tracking UI for local viewing of run results. To launch a production
     server, use the "mlflow server" command instead.
 
-    The UI will be visible at http://localhost:5000 by default, and only accept connections
+    The UI will be visible at http://localhost:5000 by default, and only accepts connections
     from the local machine. To let the UI server accept connections from other machines, you will
     need to pass ``--host 0.0.0.0`` to listen on all network interfaces (or a specific interface
     address).
@@ -263,11 +284,9 @@ def ui(backend_store_uri, default_artifact_root, artifacts_destination, port, ho
     if not backend_store_uri:
         backend_store_uri = DEFAULT_LOCAL_FILE_AND_ARTIFACT_PATH
 
-    if not default_artifact_root:
-        if is_local_uri(backend_store_uri):
-            default_artifact_root = backend_store_uri
-        else:
-            default_artifact_root = DEFAULT_LOCAL_FILE_AND_ARTIFACT_PATH
+    default_artifact_root = resolve_default_artifact_root(
+        serve_artifacts, default_artifact_root, backend_store_uri, resolve_to_local=True
+    )
 
     try:
         initialize_backend_stores(backend_store_uri, default_artifact_root)
@@ -279,7 +298,15 @@ def ui(backend_store_uri, default_artifact_root, artifacts_destination, port, ho
     # TODO: We eventually want to disable the write path in this version of the server.
     try:
         _run_server(
-            backend_store_uri, default_artifact_root, artifacts_destination, host, port, None, 1
+            backend_store_uri,
+            default_artifact_root,
+            serve_artifacts,
+            False,
+            artifacts_destination,
+            host,
+            port,
+            None,
+            1,
         )
     except ShellCommandException:
         eprint("Running the mlflow server failed. Please see the logs above for details.")
@@ -315,10 +342,24 @@ def _validate_static_prefix(ctx, param, value):  # pylint: disable=unused-argume
     "--default-artifact-root",
     metavar="URI",
     default=None,
-    help="Local or S3 URI to store artifacts, for new experiments. "
-    "Note that this flag does not impact already-created experiments. "
-    "Default: Within file store, if a file:/ URI is provided. If a sql backend is"
-    " used, then this option is required.",
+    help="Directory in which to store artifacts for any new experiments created. For tracking "
+    "server backends that rely on SQL, this option is required in order to store artifacts. "
+    "Note that this flag does not impact already-created experiments with any previous "
+    "configuration of an MLflow server instance. "
+    f"By default, data will be logged to the {DEFAULT_ARTIFACTS_URI} uri proxy if "
+    "the --serve-artifacts option is enabled. Otherwise, the default location will "
+    f"be {DEFAULT_LOCAL_FILE_AND_ARTIFACT_PATH}.",
+)
+@cli_args.SERVE_ARTIFACTS
+@click.option(
+    "--artifacts-only",
+    is_flag=True,
+    default=False,
+    help="If specified, configures the mlflow server to be used only for proxied artifact serving. "
+    "With this mode enabled, functionality of the mlflow tracking service (e.g. run creation, "
+    "metric logging, and parameter logging) is disabled. The server will only expose "
+    "endpoints for uploading, downloading, and listing artifacts. "
+    "Default: False",
 )
 @cli_args.ARTIFACTS_DESTINATION
 @cli_args.HOST
@@ -348,6 +389,8 @@ def _validate_static_prefix(ctx, param, value):  # pylint: disable=unused-argume
 def server(
     backend_store_uri,
     default_artifact_root,
+    serve_artifacts,
+    artifacts_only,
     artifacts_destination,
     host,
     port,
@@ -360,7 +403,7 @@ def server(
     """
     Run the MLflow tracking server.
 
-    The server which listen on http://localhost:5000 by default, and only accept connections
+    The server listens on http://localhost:5000 by default and only accepts connections
     from the local machine. To let the server accept connections from other machines, you will need
     to pass ``--host 0.0.0.0`` to listen on all network interfaces
     (or a specific interface address).
@@ -374,15 +417,10 @@ def server(
     if not backend_store_uri:
         backend_store_uri = DEFAULT_LOCAL_FILE_AND_ARTIFACT_PATH
 
-    if not default_artifact_root:
-        if is_local_uri(backend_store_uri):
-            default_artifact_root = backend_store_uri
-        else:
-            eprint(
-                "Option 'default-artifact-root' is required, when backend store is not "
-                "local file based."
-            )
-            sys.exit(1)
+    default_artifact_root = resolve_default_artifact_root(
+        serve_artifacts, default_artifact_root, backend_store_uri
+    )
+    artifacts_only_config_validation(artifacts_only, backend_store_uri)
 
     try:
         initialize_backend_stores(backend_store_uri, default_artifact_root)
@@ -395,6 +433,8 @@ def server(
         _run_server(
             backend_store_uri,
             default_artifact_root,
+            serve_artifacts,
+            artifacts_only,
             artifacts_destination,
             host,
             port,
@@ -427,7 +467,6 @@ def server(
     " are not specified, data is removed for all runs in the `deleted`"
     " lifecycle stage.",
 )
-@experimental
 def gc(backend_store_uri, run_ids):
     """
     Permanently delete runs in the `deleted` lifecycle stage from the specified backend store.
@@ -453,7 +492,7 @@ def gc(backend_store_uri, run_ids):
         artifact_repo = get_artifact_repository(run.info.artifact_uri)
         artifact_repo.delete_artifacts()
         backend_store._hard_delete_run(run_id)
-        print("Run with ID %s has been permanently deleted." % str(run_id))
+        click.echo("Run with ID %s has been permanently deleted." % str(run_id))
 
 
 cli.add_command(mlflow.deployments.cli.commands)
@@ -461,6 +500,7 @@ cli.add_command(mlflow.experiments.commands)
 cli.add_command(mlflow.store.artifact.cli.commands)
 cli.add_command(mlflow.runs.commands)
 cli.add_command(mlflow.db.commands)
+cli.add_command(mlflow.pipelines.cli.commands)
 
 try:
     # pylint: disable=unused-import
