@@ -10,6 +10,7 @@ import signal
 import shutil
 from subprocess import check_call, Popen
 import sys
+import logging
 
 from pkg_resources import resource_filename
 
@@ -20,7 +21,10 @@ from mlflow import pyfunc, mleap
 from mlflow.models import Model
 from mlflow.models.model import MLMODEL_FILE_NAME
 from mlflow.models.docker_utils import DISABLE_ENV_CREATION
+from mlflow.pyfunc import scoring_server, mlserver
 from mlflow.version import VERSION as MLFLOW_VERSION
+from mlflow.utils import env_manager as em
+from mlflow.utils.virtualenv import _get_or_create_virtualenv
 
 MODEL_PATH = "/opt/ml/model"
 
@@ -28,20 +32,29 @@ MODEL_PATH = "/opt/ml/model"
 DEPLOYMENT_CONFIG_KEY_FLAVOR_NAME = "MLFLOW_DEPLOYMENT_FLAVOR_NAME"
 
 DEFAULT_SAGEMAKER_SERVER_PORT = 8080
+DEFAULT_INFERENCE_SERVER_PORT = 8000
+DEFAULT_NGINX_SERVER_PORT = 8080
+DEFAULT_MLSERVER_PORT = 8080
 
 SUPPORTED_FLAVORS = [pyfunc.FLAVOR_NAME, mleap.FLAVOR_NAME]
 
 DISABLE_NGINX = "DISABLE_NGINX"
+ENABLE_MLSERVER = "ENABLE_MLSERVER"
+
+SERVING_ENVIRONMENT = "SERVING_ENVIRONMENT"
 
 
-def _init(cmd):
+_logger = logging.getLogger(__name__)
+
+
+def _init(cmd, env_manager):
     """
     Initialize the container and execute command.
 
     :param cmd: Command param passed by Sagemaker. Can be  "serve" or "train" (unimplemented).
     """
     if cmd == "serve":
-        _serve()
+        _serve(env_manager)
     elif cmd == "train":
         _train()
     else:
@@ -50,7 +63,7 @@ def _init(cmd):
         )
 
 
-def _serve():
+def _serve(env_manager):
     """
     Serve the model.
 
@@ -68,18 +81,20 @@ def _serve():
     if serving_flavor == mleap.FLAVOR_NAME:
         _serve_mleap()
     elif pyfunc.FLAVOR_NAME in m.flavors:
-        _serve_pyfunc(m)
+        _serve_pyfunc(m, env_manager)
     else:
         raise Exception("This container only supports models with the MLeap or PyFunc flavors.")
 
 
-def _install_pyfunc_deps(model_path=None, install_mlflow=False):
+def _install_pyfunc_deps(
+    model_path=None, install_mlflow=False, enable_mlserver=False, env_manager=em.CONDA
+):
     """
     Creates a conda env for serving the model at the specified path and installs almost all serving
     dependencies into the environment - MLflow is not installed as it's not available via conda.
     """
     # If model is a pyfunc model, create its conda env (even if it also has mleap flavor)
-    has_env = False
+    activate_cmd = []
     if model_path:
         model_config_path = os.path.join(model_path, MLMODEL_FILE_NAME)
         model = Model.load(model_config_path)
@@ -89,24 +104,35 @@ def _install_pyfunc_deps(model_path=None, install_mlflow=False):
             return
         conf = model.flavors[pyfunc.FLAVOR_NAME]
         if pyfunc.ENV in conf:
-            print("creating and activating custom environment")
+            _logger.info("creating and activating custom environment")
             env = conf[pyfunc.ENV]
             env_path_dst = os.path.join("/opt/mlflow/", env)
             env_path_dst_dir = os.path.dirname(env_path_dst)
             if not os.path.exists(env_path_dst_dir):
                 os.makedirs(env_path_dst_dir)
             shutil.copyfile(os.path.join(MODEL_PATH, env), env_path_dst)
-            conda_create_model_env = "conda env create -n custom_env -f {}".format(env_path_dst)
-            if Popen(["bash", "-c", conda_create_model_env]).wait() != 0:
-                raise Exception("Failed to create model environment.")
-            has_env = True
-    activate_cmd = ["source /miniconda/bin/activate custom_env"] if has_env else []
+            if env_manager == em.CONDA:
+                conda_create_model_env = "conda env create -n custom_env -f {}".format(env_path_dst)
+                if Popen(["bash", "-c", conda_create_model_env]).wait() != 0:
+                    raise Exception("Failed to create model environment.")
+                activate_cmd = ["source /miniconda/bin/activate custom_env"]
+            elif env_manager == em.VIRTUALENV:
+                env_activate_cmd = _get_or_create_virtualenv(model_path)
+                path = env_activate_cmd.split(" ")[-1]
+                os.symlink(path, "/opt/activate")
+                activate_cmd = [env_activate_cmd]
+
     # NB: install gunicorn[gevent] from pip rather than from conda because gunicorn is already
     # dependency of mlflow on pip and we expect mlflow to be part of the environment.
-    install_server_deps = ["pip install gunicorn[gevent]"]
+    server_deps = ["gunicorn[gevent]"]
+    if enable_mlserver:
+        server_deps = ["mlserver", "mlserver-mlflow"]
+
+    install_server_deps = [f"pip install {' '.join(server_deps)}"]
     if Popen(["bash", "-c", " && ".join(activate_cmd + install_server_deps)]).wait() != 0:
         raise Exception("Failed to install serving dependencies into the model environment.")
-    if has_env and install_mlflow:
+
+    if len(activate_cmd) and install_mlflow:
         install_mlflow_cmd = [
             "pip install /opt/mlflow/."
             if _container_includes_mlflow_source()
@@ -114,40 +140,66 @@ def _install_pyfunc_deps(model_path=None, install_mlflow=False):
         ]
         if Popen(["bash", "-c", " && ".join(activate_cmd + install_mlflow_cmd)]).wait() != 0:
             raise Exception("Failed to install mlflow into the model environment.")
+    return activate_cmd
 
 
-def _serve_pyfunc(model):
+def _serve_pyfunc(model, env_manager):
+    # option to disable manually nginx. The default behavior is to enable nginx.
+    disable_nginx = os.getenv(DISABLE_NGINX, "false").lower() == "true"
+    enable_mlserver = os.getenv(ENABLE_MLSERVER, "false").lower() == "true"
+    disable_env_creation = os.environ.get(DISABLE_ENV_CREATION) == "true"
+
     conf = model.flavors[pyfunc.FLAVOR_NAME]
     bash_cmds = []
     if pyfunc.ENV in conf:
-        if not os.environ.get(DISABLE_ENV_CREATION) == "true":
-            _install_pyfunc_deps(MODEL_PATH, install_mlflow=True)
-        bash_cmds += ["source /miniconda/bin/activate custom_env"]
-    nginx_conf = resource_filename(mlflow.models.__name__, "container/scoring_server/nginx.conf")
+        if not disable_env_creation:
+            _install_pyfunc_deps(
+                MODEL_PATH,
+                install_mlflow=True,
+                enable_mlserver=enable_mlserver,
+                env_manager=env_manager,
+            )
+        bash_cmds += (
+            ["source /miniconda/bin/activate custom_env"]
+            if env_manager == em.CONDA
+            else ["source /opt/activate"]
+        )
 
-    # option to disable manually nginx. The default behavior is to enable nginx.
-    start_nginx = False if os.getenv(DISABLE_NGINX, "false").lower() == "true" else True
-    nginx = Popen(["nginx", "-c", nginx_conf]) if start_nginx else None
+    procs = []
 
-    # link the log streams to stdout/err so they will be logged to the container logs.
-    # Default behavior is to do the redirection unless explicitly specified by environment variable.
+    start_nginx = True
+    if disable_nginx or enable_mlserver:
+        start_nginx = False
 
     if start_nginx:
+        nginx_conf = resource_filename(
+            mlflow.models.__name__, "container/scoring_server/nginx.conf"
+        )
+
+        nginx = Popen(["nginx", "-c", nginx_conf]) if start_nginx else None
+
+        # link the log streams to stdout/err so they will be logged to the container logs.
+        # Default behavior is to do the redirection unless explicitly specified
+        # by environment variable.
         check_call(["ln", "-sf", "/dev/stdout", "/var/log/nginx/access.log"])
         check_call(["ln", "-sf", "/dev/stderr", "/var/log/nginx/error.log"])
 
+        procs.append(nginx)
+
     cpu_count = multiprocessing.cpu_count()
     os.system("pip -V")
     os.system("python -V")
     os.system('python -c"from mlflow.version import VERSION as V; print(V)"')
-    cmd = (
-        "gunicorn -w {cpu_count} ".format(cpu_count=cpu_count)
-        + "${GUNICORN_CMD_ARGS} mlflow.models.container.scoring_server.wsgi:app"
-    )
-    bash_cmds.append(cmd)
-    gunicorn = Popen(["/bin/bash", "-c", " && ".join(bash_cmds)])
 
-    procs = [p for p in [nginx, gunicorn] if p]
+    inference_server = mlserver if enable_mlserver else scoring_server
+    # Since MLServer will run without NGINX, expose the server in the `8080`
+    # port, which is the assumed "public" port.
+    port = DEFAULT_MLSERVER_PORT if enable_mlserver else DEFAULT_INFERENCE_SERVER_PORT
+    cmd, cmd_env = inference_server.get_cmd(model_uri=MODEL_PATH, nworkers=cpu_count, port=port)
+
+    bash_cmds.append(cmd)
+    inference_server_process = Popen(["/bin/bash", "-c", " && ".join(bash_cmds)], env=cmd_env)
+    procs.append(inference_server_process)
 
     signal.signal(signal.SIGTERM, lambda a, b: _sigterm_handler(pids=[p.pid for p in procs]))
     # If either subprocess exits, so do we.
@@ -197,7 +249,7 @@ def _sigterm_handler(pids):
     Attempt to kill all launched processes and exit.
 
     """
-    print("Got sigterm signal, exiting.")
+    _logger.info("Got sigterm signal, exiting.")
     for pid in pids:
         try:
             os.kill(pid, signal.SIGTERM)
