@@ -7,6 +7,7 @@ import textwrap
 import time
 import logging
 import posixpath
+import re
 
 from shlex import quote as shlex_quote
 
@@ -42,12 +43,26 @@ DBFS_EXPERIMENT_DIR_BASE = "mlflow-experiments"
 
 _logger = logging.getLogger(__name__)
 
+_MLFLOW_GIT_URI_REGEX = re.compile(r"^git\+https://github.com/[\w-]+/mlflow")
+
+
+def _is_mlflow_git_uri(s):
+    return bool(_MLFLOW_GIT_URI_REGEX.match(s))
+
+
+def _contains_mlflow_git_uri(libraries):
+    for lib in libraries:
+        package = lib.get("pypi", {}).get("package")
+        if package and _is_mlflow_git_uri(package):
+            return True
+    return False
+
 
 def before_run_validations(tracking_uri, backend_config):
     """Validations to perform before running a project on Databricks."""
     if backend_config is None:
         raise ExecutionException(
-            "Backend spec must be provided when launching MLflow project " "runs on Databricks."
+            "Backend spec must be provided when launching MLflow project runs on Databricks."
         )
     elif "existing_cluster_id" in backend_config:
         raise MlflowException(
@@ -69,7 +84,7 @@ def before_run_validations(tracking_uri, backend_config):
         )
 
 
-class DatabricksJobRunner(object):
+class DatabricksJobRunner:
     """
     Helper class for running an MLflow project as a Databricks Job.
     :param databricks_profile: Optional Databricks CLI profile to use to fetch hostname &
@@ -122,7 +137,7 @@ class DatabricksJobRunner(object):
         )
         try:
             json_response_obj = json.loads(response.text)
-        except Exception:  # pylint: disable=broad-except
+        except Exception:
             raise MlflowException(
                 "API request to check existence of file at DBFS path %s failed with status code "
                 "%s. Response body: %s" % (dbfs_path, response.status_code, response.text)
@@ -153,6 +168,11 @@ class DatabricksJobRunner(object):
             return None if os.path.basename(x.name) == "mlruns" else x
 
         try:
+            directory_size = file_utils._get_local_project_dir_size(project_dir)
+            _logger.info(
+                f"=== Creating tarball from {project_dir} in temp directory {temp_tarfile_dir} ==="
+            )
+            _logger.info(f"=== Total file size to compress: {directory_size} KB ===")
             file_utils.make_tarfile(
                 temp_tar_filename, project_dir, DB_TARFILE_ARCHIVE_NAME, custom_filter=custom_filter
             )
@@ -165,8 +185,12 @@ class DatabricksJobRunner(object):
                 "projects-code",
                 "%s.tar.gz" % tarfile_hash,
             )
+            tar_size = file_utils._get_local_file_size(temp_tar_filename)
             dbfs_fuse_uri = posixpath.join("/dbfs", dbfs_path)
             if not self._dbfs_path_exists(dbfs_path):
+                _logger.info(
+                    f"=== Uploading project tarball (size: {tar_size} KB) to {dbfs_fuse_uri} ==="
+                )
                 self._upload_to_dbfs(temp_tar_filename, dbfs_fuse_uri)
                 _logger.info("=== Finished uploading project to %s ===", dbfs_fuse_uri)
             else:
@@ -194,26 +218,35 @@ class DatabricksJobRunner(object):
                  `Runs Get <https://docs.databricks.com/api/latest/jobs.html#runs-get>`_ API.
         """
         if is_release_version():
-            libraries = [{"pypi": {"package": "mlflow==%s" % VERSION}}]
+            mlflow_lib = {"pypi": {"package": "mlflow==%s" % VERSION}}
         else:
             # When running a non-release version as the client the same version will not be
             # available within Databricks.
             _logger.warning(
                 (
                     "Your client is running a non-release version of MLFlow. "
-                    "This version is not avaialable on the databricks runtime. "
+                    "This version is not available on the databricks runtime. "
                     "MLFlow will fallback the MLFlow version provided by the runtime. "
                     "This might lead to unforeseen issues. "
                 )
             )
-            libraries = [{"pypi": {"package": "'mlflow<=%s'" % VERSION}}]
+            mlflow_lib = {"pypi": {"package": "'mlflow<=%s'" % VERSION}}
 
         # Check syntax of JSON - if it contains libraries and new_cluster, pull those out
         if "new_cluster" in cluster_spec:
             # Libraries are optional, so we don't require that this be specified
-            if "libraries" in cluster_spec:
-                libraries.extend(cluster_spec["libraries"])
+            cluster_spec_libraries = cluster_spec.get("libraries", [])
+            libraries = (
+                # This is for development purposes only. If the cluster spec already includes
+                # an MLflow Git URI, then we don't append `mlflow_lib` to avoid having
+                # two different pip requirements for mlflow.
+                cluster_spec_libraries
+                if _contains_mlflow_git_uri(cluster_spec_libraries)
+                else cluster_spec_libraries + [mlflow_lib]
+            )
             cluster_spec = cluster_spec["new_cluster"]
+        else:
+            libraries = [mlflow_lib]
 
         # Make jobs API request to launch run.
         req_body_json = {
@@ -222,12 +255,21 @@ class DatabricksJobRunner(object):
             "shell_command_task": {"command": command, "env_vars": env_vars},
             "libraries": libraries,
         }
+        _logger.info("=== Submitting a run to execute the MLflow project... ===")
         run_submit_res = self._jobs_runs_submit(req_body_json)
         databricks_run_id = run_submit_res["run_id"]
         return databricks_run_id
 
     def run_databricks(
-        self, uri, entry_point, work_dir, parameters, experiment_id, cluster_spec, run_id
+        self,
+        uri,
+        entry_point,
+        work_dir,
+        parameters,
+        experiment_id,
+        cluster_spec,
+        run_id,
+        env_manager,
     ):
         tracking_uri = _get_tracking_uri_for_run()
         dbfs_fuse_uri = self._upload_project_to_dbfs(work_dir, experiment_id)
@@ -237,7 +279,9 @@ class DatabricksJobRunner(object):
         }
         _logger.info("=== Running entry point %s of project %s on Databricks ===", entry_point, uri)
         # Launch run on Databricks
-        command = _get_databricks_run_cmd(dbfs_fuse_uri, run_id, entry_point, parameters)
+        command = _get_databricks_run_cmd(
+            dbfs_fuse_uri, run_id, entry_point, parameters, env_manager
+        )
         return self._run_shell_command_job(uri, command, env_vars, cluster_spec)
 
     def _get_status(self, databricks_run_id):
@@ -283,10 +327,17 @@ def _get_tracking_uri_for_run():
     return uri
 
 
-def _get_cluster_mlflow_run_cmd(project_dir, run_id, entry_point, parameters):
-    mlflow_run_arr = list(
-        map(shlex_quote, ["mlflow", "run", project_dir, "--entry-point", entry_point])
-    )
+def _get_cluster_mlflow_run_cmd(project_dir, run_id, entry_point, parameters, env_manager):
+    cmd = [
+        "mlflow",
+        "run",
+        project_dir,
+        "--entry-point",
+        entry_point,
+    ]
+    if env_manager:
+        cmd += ["--env-manager", env_manager]
+    mlflow_run_arr = list(map(shlex_quote, cmd))
     if run_id:
         mlflow_run_arr.extend(["-c", json.dumps({MLFLOW_LOCAL_BACKEND_RUN_ID_CONFIG: run_id})])
     if parameters:
@@ -295,7 +346,7 @@ def _get_cluster_mlflow_run_cmd(project_dir, run_id, entry_point, parameters):
     return mlflow_run_arr
 
 
-def _get_databricks_run_cmd(dbfs_fuse_tar_uri, run_id, entry_point, parameters):
+def _get_databricks_run_cmd(dbfs_fuse_tar_uri, run_id, entry_point, parameters, env_manager):
     """
     Generate MLflow CLI command to run on Databricks cluster in order to launch a run on Databricks.
     """
@@ -306,7 +357,13 @@ def _get_databricks_run_cmd(dbfs_fuse_tar_uri, run_id, entry_point, parameters):
     )
     project_dir = posixpath.join(DB_PROJECTS_BASE, tar_hash)
 
-    mlflow_run_arr = _get_cluster_mlflow_run_cmd(project_dir, run_id, entry_point, parameters)
+    mlflow_run_arr = _get_cluster_mlflow_run_cmd(
+        project_dir,
+        run_id,
+        entry_point,
+        parameters,
+        env_manager,
+    )
     mlflow_run_cmd = " ".join([shlex_quote(elem) for elem in mlflow_run_arr])
     shell_command = textwrap.dedent(
         """
@@ -336,7 +393,9 @@ def _get_databricks_run_cmd(dbfs_fuse_tar_uri, run_id, entry_point, parameters):
     return ["bash", "-c", shell_command]
 
 
-def run_databricks(remote_run, uri, entry_point, work_dir, parameters, experiment_id, cluster_spec):
+def run_databricks(
+    remote_run, uri, entry_point, work_dir, parameters, experiment_id, cluster_spec, env_manager
+):
     """
     Run the project at the specified URI on Databricks, returning a ``SubmittedRun`` that can be
     used to query the run's status or wait for the resulting Databricks Job run to terminate.
@@ -344,7 +403,7 @@ def run_databricks(remote_run, uri, entry_point, work_dir, parameters, experimen
     run_id = remote_run.info.run_id
     db_job_runner = DatabricksJobRunner(databricks_profile_uri=tracking.get_tracking_uri())
     db_run_id = db_job_runner.run_databricks(
-        uri, entry_point, work_dir, parameters, experiment_id, cluster_spec, run_id
+        uri, entry_point, work_dir, parameters, experiment_id, cluster_spec, run_id, env_manager
     )
     submitted_run = DatabricksSubmittedRun(db_run_id, run_id, db_job_runner)
     submitted_run._print_description_and_log_tags()
