@@ -4,15 +4,16 @@ import itertools
 import functools
 import os
 import uuid
-import warnings
 from abc import abstractmethod
 from contextlib import contextmanager
+import typing
 
 import mlflow
 import mlflow.utils.autologging_utils
 from mlflow.entities.run_status import RunStatus
 from mlflow.tracking.client import MlflowClient
 from mlflow.utils import gorilla
+from mlflow.utils import is_iterator
 from mlflow.utils.autologging_utils import _logger
 from mlflow.utils.autologging_utils.events import AutologgingEventLogger
 from mlflow.utils.autologging_utils.logging_and_warnings import (
@@ -26,28 +27,20 @@ _AUTOLOGGING_TEST_MODE_ENV_VAR = "MLFLOW_AUTOLOGGING_TESTING"
 _AUTOLOGGING_PATCHES = {}
 
 
-def try_mlflow_log(fn, *args, **kwargs):
-    """
-    Catch exceptions and log a warning to avoid autolog throwing.
-    """
-    try:
-        return fn(*args, **kwargs)
-    except Exception as e:
-        if is_testing():
-            raise
-        else:
-            warnings.warn("Logging to MLflow failed: " + str(e), stacklevel=2)
-
-
 # Function attribute used for testing purposes to verify that a given function
-# has been wrapped with the `exception_safe_function` decorator
+# has been wrapped with the `exception_safe_function_for_class` and
+# `picklable_exception_safe_function` decorators
 _ATTRIBUTE_EXCEPTION_SAFE = "exception_safe"
 
 
-def exception_safe_function(function):
+def exception_safe_function_for_class(function):
     """
     Wraps the specified function with broad exception handling to guard
     against unexpected errors during autologging.
+    Note this function creates an unpicklable function as `safe_function` is locally defined,
+    but a class instance containing methods decorated by this function should be pickalable,
+    because pickle only saves instance attributes, not methods.
+    See https://docs.python.org/3/library/pickle.html#pickling-class-instances for more details.
     """
     if is_testing():
         setattr(function, _ATTRIBUTE_EXCEPTION_SAFE, True)
@@ -65,6 +58,27 @@ def exception_safe_function(function):
     return safe_function
 
 
+def _safe_function(function, *args, **kwargs):
+    try:
+        return function(*args, **kwargs)
+    except Exception as e:
+        if is_testing():
+            raise
+        else:
+            _logger.warning("Encountered unexpected error during autologging: %s", e)
+
+
+def picklable_exception_safe_function(function):
+    """
+    Wraps the specified function with broad exception handling to guard
+    against unexpected errors during autologging while preserving picklability.
+    """
+    if is_testing():
+        setattr(function, _ATTRIBUTE_EXCEPTION_SAFE, True)
+
+    return update_wrapper_extended(functools.partial(_safe_function, function), function)
+
+
 def _exception_safe_class_factory(base_class):
     """
     Creates an exception safe metaclass that inherits from `base_class`.
@@ -89,7 +103,7 @@ def _exception_safe_class_factory(base_class):
             for m in dct:
                 # class methods or static methods are not callable.
                 if callable(dct[m]):
-                    dct[m] = exception_safe_function(dct[m])
+                    dct[m] = exception_safe_function_for_class(dct[m])
             return base_class.__new__(cls, name, bases, dct)
 
     return _ExceptionSafeClass
@@ -208,26 +222,24 @@ def with_managed_run(autologging_integration, patch_function, tags=None):
 
         class PatchWithManagedRun(patch_function):
             def __init__(self):
-                super(PatchWithManagedRun, self).__init__()
+                super().__init__()
                 self.managed_run = None
 
             def _patch_implementation(self, original, *args, **kwargs):
                 if not mlflow.active_run():
-                    self.managed_run = try_mlflow_log(create_managed_run)
+                    self.managed_run = create_managed_run()
 
-                result = super(PatchWithManagedRun, self)._patch_implementation(
-                    original, *args, **kwargs
-                )
+                result = super()._patch_implementation(original, *args, **kwargs)
 
                 if self.managed_run:
-                    try_mlflow_log(mlflow.end_run, RunStatus.to_string(RunStatus.FINISHED))
+                    mlflow.end_run(RunStatus.to_string(RunStatus.FINISHED))
 
                 return result
 
             def _on_exception(self, e):
                 if self.managed_run:
-                    try_mlflow_log(mlflow.end_run, RunStatus.to_string(RunStatus.FAILED))
-                super(PatchWithManagedRun, self)._on_exception(e)
+                    mlflow.end_run(RunStatus.to_string(RunStatus.FAILED))
+                super()._on_exception(e)
 
         return PatchWithManagedRun
 
@@ -236,7 +248,7 @@ def with_managed_run(autologging_integration, patch_function, tags=None):
         def patch_with_managed_run(original, *args, **kwargs):
             managed_run = None
             if not mlflow.active_run():
-                managed_run = try_mlflow_log(create_managed_run)
+                managed_run = create_managed_run()
 
             try:
                 result = patch_function(original, *args, **kwargs)
@@ -245,11 +257,11 @@ def with_managed_run(autologging_integration, patch_function, tags=None):
                 # that runs are terminated if a user prematurely interrupts training execution
                 # (e.g. via sigint / ctrl-c)
                 if managed_run:
-                    try_mlflow_log(mlflow.end_run, RunStatus.to_string(RunStatus.FAILED))
+                    mlflow.end_run(RunStatus.to_string(RunStatus.FAILED))
                 raise
             else:
                 if managed_run:
-                    try_mlflow_log(mlflow.end_run, RunStatus.to_string(RunStatus.FINISHED))
+                    mlflow.end_run(RunStatus.to_string(RunStatus.FINISHED))
                 return result
 
         return patch_with_managed_run
@@ -417,7 +429,8 @@ def safe_patch(
                 # warning behavior during original function execution, since autologging is being
                 # skipped
                 with set_non_mlflow_warnings_behavior_for_current_thread(
-                    disable_warnings=False, reroute_warnings=False,
+                    disable_warnings=False,
+                    reroute_warnings=False,
                 ):
                     return original(*args, **kwargs)
 
@@ -440,7 +453,9 @@ def safe_patch(
                     log_fn(*args)
                 except Exception as e:
                     _logger.debug(
-                        "Failed to log autologging event via '%s'. Exception: %s", log_fn, e,
+                        "Failed to log autologging event via '%s'. Exception: %s",
+                        log_fn,
+                        e,
                     )
 
             def call_original_fn_with_event_logging(original_fn, og_args, og_kwargs):
@@ -485,7 +500,14 @@ def safe_patch(
                     def call_original(*og_args, **og_kwargs):
                         def _original_fn(*_og_args, **_og_kwargs):
                             if is_testing():
-                                _validate_args(args, kwargs, og_args, og_kwargs)
+                                _validate_args(
+                                    autologging_integration,
+                                    function_name,
+                                    args,
+                                    kwargs,
+                                    og_args,
+                                    og_kwargs,
+                                )
                                 # By the time `original` is called by the patch implementation, we
                                 # assume that either: 1. the patch implementation has already
                                 # created an MLflow run or 2. the patch code will not create an
@@ -505,7 +527,8 @@ def safe_patch(
                             # (`silent=True`), since these warnings originate from the ML framework
                             # or one of its dependencies and are likely relevant to the caller
                             with set_non_mlflow_warnings_behavior_for_current_thread(
-                                disable_warnings=False, reroute_warnings=False,
+                                disable_warnings=False,
+                                reroute_warnings=False,
                             ):
                                 original_result = original(*_og_args, **_og_kwargs)
                                 return original_result
@@ -636,7 +659,7 @@ def revert_patches(autologging_integration):
 
     :param autologging_integration: The name of the autologging integration associated with the
                                     patch. Note: If called via fluent api
-                                    (`autologging_integration="mlfow"`), then revert all patches
+                                    (`autologging_integration="mlflow"`), then revert all patches
                                     for all active autologging integrations.
     """
     for patch in _AUTOLOGGING_PATCHES.get(autologging_integration, []):
@@ -757,8 +780,110 @@ def _validate_autologging_run(autologging_integration, run_id):
     ), "Autologging run with id {} has a non-terminal status '{}'".format(run_id, run.info.status)
 
 
+class ValidationExemptArgument(typing.NamedTuple):
+    """
+    A NamedTuple representing the properties of an argument that is exempt from validation
+
+    autologging_integration: The name of the autologging integration.
+    function_name: The name of the function that is being validated.
+    type_function: A Callable that accepts an object and returns True if the given object matches
+                   the argument type. Returns False otherwise.
+    positional_argument_index: The index of the argument in the function signature.
+    keyword_argument_name: The name of the argument in the function signature.
+    """
+
+    autologging_integration: str
+    function_name: str
+    type_function: typing.Callable
+    positional_argument_index: int = None
+    keyword_argument_name: str = None
+
+    def matches(
+        self,
+        autologging_integration,
+        function_name,
+        value,
+        argument_index=None,
+        argument_name=None,
+    ):
+        """
+        This method checks if the properties provided through the function arguments matches the
+        properties defined in the NamedTuple.
+
+        :param autologging_integration: The name of an autologging integration.
+        :param function_name: The name of the function that is being matched.
+        :param value: The value of the argument.
+        :param argument_index: The index of the argument, if it is passed as a positional
+                                          argument. Otherwise it is None.
+        :param argument_name: The name of the argument, if it is passed as a keyword
+                                      argument. Otherwise it is None.
+        :return: Returns True if the given function properties matches the exempt argument's
+                 properties. Returns False otherwise.
+        """
+        return (
+            self.autologging_integration == autologging_integration
+            and self.function_name == function_name
+            and (
+                self.positional_argument_index == argument_index
+                or self.keyword_argument_name == argument_name
+            )
+            and self.type_function(value)
+        )
+
+
+# WARNING: Exemptions should NOT be introduced unless absolutely necessary. If deemed necessary,
+#          clear reasons must be provided as comment in addition to thorough integration tests.
+_VALIDATION_EXEMPT_ARGUMENTS = [
+    # When extracting implicitly defined `batch_size` in the case that `x` is a generator or a
+    # generator class, we need to consume and restore the first element back to the generator to
+    # calculate the `batch_size`. This means that:
+    # 1. The type of `x` will become 'generator' regardless if user provided `x` as a generator or a
+    #    custom generator class.
+    # 2. The instance of `x` will be different, since we reconstructed the generator after consuming
+    #    the first element.
+    ValidationExemptArgument("tensorflow", "fit", is_iterator, 1, "x")
+]
+
+
+def _is_arg_exempt_from_validation(
+    autologging_integration,
+    function_name,
+    argument,
+    argument_index=None,
+    argument_name=None,
+):
+    """
+    This function is responsible for determining whether or not an argument is exempt from autolog
+    safety validations. This includes both type checking and immutable checking.
+
+    :param autologging_integration: The name of the autologging integration.
+    :param function_name: The name of the function that is being validated.
+    :param argument: The actual argument.
+    :param argument_index: The index of the argument, if it is passed as a positional
+                                      argument. Otherwise it is None.
+    :param argument_name: The name of the argument, if it is passed as a keyword argument.
+                                  Otherwise it is None.
+    :return: True or False
+    """
+    return any(
+        exemption.matches(
+            autologging_integration,
+            function_name,
+            argument,
+            argument_index,
+            argument_name,
+        )
+        for exemption in _VALIDATION_EXEMPT_ARGUMENTS
+    )
+
+
 def _validate_args(
-    user_call_args, user_call_kwargs, autologging_call_args, autologging_call_kwargs
+    autologging_integration,
+    function_name,
+    user_call_args,
+    user_call_kwargs,
+    autologging_call_args,
+    autologging_call_kwargs,
 ):
     """
     Used for testing purposes to verify that, when a patched ML function calls its underlying
@@ -767,8 +892,9 @@ def _validate_args(
         - All arguments supplied to the patched ML function are forwarded to the
           original ML function
         - Any additional arguments supplied to the original function are exception safe (i.e.
-          they are either functions decorated with the `@exception_safe_function` decorator
-          or classes / instances of classes with type `ExceptionSafeClass`
+          they are either functions decorated with the `@exception_safe_function_for_class` or
+          `@pickalable_exception_safe_function` decorators, or classes / instances of classes with
+          type `ExceptionSafeClass`
     """
 
     def _validate_new_input(inp):
@@ -776,7 +902,8 @@ def _validate_args(
         Validates a new input (arg or kwarg) introduced to the underlying / original ML function
         call during the execution of a patched ML function. The new input is valid if:
 
-            - The new input is a function that has been decorated with `exception_safe_function`
+            - The new input is a function that has been decorated with
+              `exception_safe_function_for_class` or `pickalable_exception_safe_function`
             - OR the new input is a class with the `ExceptionSafeClass` metaclass
             - OR the new input is a list and each of its elements is valid according to the
               these criteria
@@ -787,8 +914,9 @@ def _validate_args(
         elif callable(inp):
             assert getattr(inp, _ATTRIBUTE_EXCEPTION_SAFE, False), (
                 "New function argument '{}' passed to original function is not exception-safe."
-                " Please decorate the function with `exception_safe_function`.".format(inp)
-            )
+                " Please decorate the function with `exception_safe_function` or "
+                "`pickalable_exception_safe_function`"
+            ).format(inp)
         else:
             assert hasattr(inp, "__class__") and type(inp.__class__) in [
                 ExceptionSafeClass,
@@ -796,11 +924,29 @@ def _validate_args(
             ], (
                 "Invalid new input '{}'. New args / kwargs introduced to `original` function "
                 "calls by patched code must either be functions decorated with "
-                "`exception_safe_function`, instances of classes with the `ExceptionSafeClass` "
-                "or `ExceptionSafeAbstractClass` metaclass safe or lists of such exception safe "
-                "functions / classes.".format(inp)
+                "`exception_safe_function_for_class`, instances of classes with the "
+                "`ExceptionSafeClass` or `ExceptionSafeAbstractClass` metaclass safe or lists of "
+                "such exception safe functions / classes.".format(inp)
             )
 
+    def _assert_autologging_input_positional_args_are_superset(
+        autologging_call_input, user_call_input
+    ):
+        length_difference = len(autologging_call_input) - len(user_call_input)
+        assert (
+            length_difference >= 0
+        ), "{} expected inputs are missing from the call to the original function.".format(
+            length_difference
+        )
+
+    def _assert_autologging_input_kwargs_are_superset(autologging_call_input, user_call_input):
+        assert set(user_call_input.keys()).issubset(set(autologging_call_input.keys())), (
+            "Keyword or dictionary arguments to original function omit"
+            " one or more expected keys: '{}'".format(
+                set(user_call_input.keys()) - set(autologging_call_input.keys())
+            )
+        )
+
     def _validate(autologging_call_input, user_call_input=None):
         """
         Validates that the specified `autologging_call_input` and `user_call_input`
@@ -814,7 +960,11 @@ def _validate_args(
               be equivalent to `user_call_input` or be a superset of `user_call_input`
             - for all other input types, `autologging_call_input` and `user_call_input`
               must be equivalent by reference equality or by object equality
+
+        :param autologging_call_input: call input from autologging
+        :param user_call_input: call input from user
         """
+
         if user_call_input is None and autologging_call_input is not None:
             _validate_new_input(autologging_call_input)
             return
@@ -826,10 +976,8 @@ def _validate_args(
         )
 
         if type(autologging_call_input) in [list, tuple]:
-            length_difference = len(autologging_call_input) - len(user_call_input)
-            assert length_difference >= 0, (
-                "{} expected inputs are missing from the call"
-                " to the original function.".format(length_difference)
+            _assert_autologging_input_positional_args_are_superset(
+                autologging_call_input, user_call_input
             )
             # If the autologging call input is longer than the user call input, we `zip_longest`
             # will pad the user call input with `None` values to ensure that the subsequent calls
@@ -837,12 +985,7 @@ def _validate_args(
             for a, u in itertools.zip_longest(autologging_call_input, user_call_input):
                 _validate(a, u)
         elif type(autologging_call_input) == dict:
-            assert set(user_call_input.keys()).issubset(set(autologging_call_input.keys())), (
-                "Keyword or dictionary arguments to original function omit"
-                " one or more expected keys: '{}'".format(
-                    set(user_call_input.keys()) - set(autologging_call_input.keys())
-                )
-            )
+            _assert_autologging_input_kwargs_are_superset(autologging_call_input, user_call_input)
             for key in autologging_call_input.keys():
                 _validate(autologging_call_input[key], user_call_input.get(key, None))
         else:
@@ -854,15 +997,41 @@ def _validate_args(
                 " Original: '{}'. Expected: '{}'".format(autologging_call_input, user_call_input)
             )
 
-    _validate(autologging_call_args, user_call_args)
-    _validate(autologging_call_kwargs, user_call_kwargs)
+    # Similar validation logic found in _validate, unraveling the list of arguments to exclude
+    # checks for any validation exempt positional arguments.
+    _assert_autologging_input_positional_args_are_superset(autologging_call_args, user_call_args)
+    for index, autologging_call_arg, user_call_arg in itertools.zip_longest(
+        range(len(user_call_args)), autologging_call_args, user_call_args
+    ):
+        if not _is_arg_exempt_from_validation(
+            autologging_integration,
+            function_name,
+            user_call_arg,
+            argument_index=index,
+        ):
+            _validate(autologging_call_arg, user_call_arg)
+
+    # Similar validation logic found in _validate, unraveling the dictionary of arguments to exclude
+    # checks for any validation exempt keyword arguments.
+    _assert_autologging_input_kwargs_are_superset(autologging_call_kwargs, user_call_kwargs)
+    for key in autologging_call_kwargs.keys():
+        if not _is_arg_exempt_from_validation(
+            autologging_integration,
+            function_name,
+            user_call_kwargs.get(key, None),
+            argument_name=key,
+        ):
+            _validate(
+                autologging_call_kwargs[key],
+                user_call_kwargs.get(key, None),
+            )
 
 
 __all__ = [
-    "try_mlflow_log",
     "safe_patch",
     "is_testing",
-    "exception_safe_function",
+    "exception_safe_function_for_class",
+    "picklable_exception_safe_function",
     "ExceptionSafeClass",
     "ExceptionSafeAbstractClass",
     "PatchFunction",
